# driver for max chips on ROX-ECU board
from digitalio import DigitalInOut
import busio

DEBUG = False  # set to True  for verbose output


# ----------------- Max registers


class REGISTERS:
    # Output and LED configuration
    SET_OUT = 0x00
    SET_LED = 0x01

    # Fault and status information
    DOI_LEVEL = 0x02
    INTERRUPT = 0x03
    OVR_LD_CH_F = 0x04
    OPN_WIR_CH_F = 0x05
    SHT_VDD_CH_F = 0x06
    GLOBAL_ERR = 0x07

    # Configuration and control
    OPN_WR_EN = 0x08
    SHT_VDD_EN = 0x09
    CONFIG1 = 0x0A
    CONFIG2 = 0x0B
    CONFIG_DI = 0x0C
    CONFIG_DO = 0x0D
    CURR_LIM = 0x0E

    # Miscellaneous and reserved
    MASK = 0x0F

    @classmethod
    def get_registers(cls) -> list[tuple[str, int]]:
        """(name,value) pairs of all register values in the class"""
        return [
            (attr, getattr(cls, attr))
            for attr in dir(cls)
            if not attr.startswith("__") and not callable(getattr(cls, attr))
        ]

    @classmethod
    def get_name(cls, value: int) -> str | int:
        """get the name of a register from its value"""
        for attr in dir(cls):
            if not attr.startswith("__") and not callable(getattr(cls, attr)):
                if getattr(cls, attr) == value:
                    return attr
        return value


def decode_global_err(register_value: int) -> list:
    """
    Decode the GlobalErr register value into human-readable errors.

    Args:
    register_value (int): The 8-bit value of the GlobalErr register.

    Returns:
    list of str: Descriptions of the active errors.
    """
    error_descriptions = {
        7: "Watchdog Error: SPI or SYNCH watchdog timeout detected.",
        6: "Loss of Ground: Loss of ground fault detected.",
        5: "Thermal Shutdown: The device has entered thermal shutdown.",
        4: "VDD Under-Voltage: VDD has fallen below the under-voltage lock-out threshold.",
        3: "VDD Warning: VDD is below the warning threshold.",
        2: "VDD Low: VDD is significantly below the nominal operating level.",
        1: "V5 Under-Voltage: V5 has fallen below its under-voltage lock-out threshold.",
        0: "Internal Supply Voltage Low: Internal supply voltage is below threshold.",
    }
    active_errors = []

    for bit in range(8):  # Check each bit from 0 to 7
        if register_value & (1 << bit):  # If the bit is set
            active_errors.append(error_descriptions[bit])

    return active_errors


class Max14906:
    """driver for MAX14906 chips on ROX-ECU board"""

    def __init__(
        self,
        spi: busio.SPI,
        cs: DigitalInOut,
        d_pins: list[DigitalInOut],
        chip_address: int = 0,
    ):
        self.chip_address = chip_address
        self._spi = spi
        self._cs = cs

        self.d_pins = d_pins

        for p in self.d_pins:
            p.switch_to_output(False)

        self.get_global_error()  # clear the global error register

        # set leds to internal operation
        self.set_bit(REGISTERS.CONFIG1, 1, False)

        # write default values (partially)
        self.write_register(REGISTERS.SET_OUT, 0x00)
        self.write_register(REGISTERS.CONFIG_DO, 0x00)

    def set_bit(self, reg: int, bit: int, value: bool) -> bytearray:
        """set a single bit in a register"""
        data = self.read_register(reg)[1]
        data = data | (1 << bit) if value else data & ~(1 << bit)

        return self.write_register(reg, data)

    def switch_to_output(self, d_pin_nr: int, value: bool) -> None:
        """switch a D pin to output and set value"""
        self.d_pins[d_pin_nr].switch_to_output(value)

        # clear SetDi bit
        self.set_bit(REGISTERS.SET_OUT, d_pin_nr + 4, False)

    def switch_to_input(self, d_pin_nr: int) -> None:
        """switch a D pin to input"""
        self.d_pins[d_pin_nr].switch_to_input()

        # set SetDi bit
        self.set_bit(REGISTERS.SET_OUT, d_pin_nr + 4, True)

    def read_register(self, reg: int) -> bytearray:
        """single cycle read"""
        mosi_byte = (self.chip_address << 6) | (reg << 1) & 0xFE

        data_out = bytes([mosi_byte, 0x00])
        data_in = bytearray(2)

        self._cs.value = False
        self._spi.write_readinto(data_out, data_in)
        self._cs.value = True

        if DEBUG:
            print(
                f"read [{reg:02X}]({REGISTERS.get_name(reg)}) > {data_out.hex(' ')} , < {data_in.hex(' ')} ({data_in[1]:08b})"
            )

        return data_in

    def write_register(self, reg: int, data_byte: int) -> bytearray:
        """single cycle write, write a single byte to a register, returns the data read back"""

        mosi_byte = (self.chip_address << 6) | (reg << 1) | 0x1

        data_out = bytes([mosi_byte, data_byte])
        data_in = bytearray(2)

        self._cs.value = False
        self._spi.write_readinto(data_out, data_in)
        self._cs.value = True

        if DEBUG:
            print(
                f"write [{reg:02X}]({REGISTERS.get_name(reg)}) {data_out.hex(' ')} , < {data_in.hex(' ')}"
            )

        return data_in

    def get_global_error(self) -> int:
        """get the global error register"""
        return self.read_register(REGISTERS.GLOBAL_ERR)[1]

    def print_registers(self) -> None:
        """print all register values"""
        global DEBUG

        debug_bck = DEBUG
        DEBUG = True

        for name, reg in REGISTERS.get_registers():
            self.read_register(reg)

        DEBUG = debug_bck



#!/usr/bin/env python3
"""
Demonstrate basic functionality of the ICU board:

   - flash the onboard LEDs
   - cycle through the RGB LED colors on Neopixel

Note: we are using asyncio here to demonstrate loop-based programming

Copyright (c) 2024 ROX Automation - Jev Kuznetsov
"""

import asyncio

from icu_board import rgb_led, led1, led2, D_PINS, max_enable

max_enable.value = True  # enable the MAX14906 chips


async def flash_led(led, interval=0.5):
    """flash the LED at the given interval"""

    while True:
        led.value = not led.value
        await asyncio.sleep(interval)


async def switch_outputs():
    """switch a couple of D pins on and off"""
    print("switching D pins on and off")

    d5 = D_PINS[4]
    d6 = D_PINS[5]

    d5.value = True
    d6.value = False

    print(d5)
    print(d6)

    while True:
        d5.value = not d5.value
        d6.value = not d6.value
        await asyncio.sleep(1)


async def cycle_neopixel():
    """cycle through the RGB colors on the Neopixel"""
    print("cycling through the RGB colors on the Neopixel")

    colors = [
        (255, 0, 0),
        (0, 255, 0),
        (0, 0, 255),
        (255, 255, 0),
        (255, 0, 255),
        (0, 255, 255),
        (255, 255, 255),
        (0, 0, 0),
    ]

    rgb_led.brightness = 0.3

    while True:
        for color in colors:
            rgb_led.fill(color)
            await asyncio.sleep(0.3)


async def main():
    """main coro"""
    await asyncio.gather(
        flash_led(led1),  # flash led1 with default frequency
        flash_led(led2, 0.4),  # flash led2 faster
        cycle_neopixel(),  # cycle through the RGB colors
        switch_outputs(),  # switch D pins on and off
    )


# ----------- run ---------------
asyncio.run(main())



"""demonstrate can usage basics on ROX ICU board"""

import struct
import time

import canio
from icu_board import can, led1


listener = can.listen(matches=[canio.Match(0x123)], timeout=0.1)


old_bus_state = None
counter = 0

while True:
    bus_state = can.state
    if bus_state != old_bus_state:
        print(f"Bus state changed to {bus_state}")
        old_bus_state = bus_state

    message = listener.receive()
    if message is not None:
        data = message.data  # type: ignore
        print(f"received message: {data!r}")

    # Send a message with the counter value
    counter += 1
    message = canio.Message(id=0x124, data=struct.pack("<I", counter))
    print(f"> {message.data.hex(" ")}")
    can.send(message)

    led1.value = not led1.value

    time.sleep(1.0)



#!/usr/bin/env python3
"""
control ICU device remotely.

ICU functions as remote I/O device with CAN interface.

Copyright (c) 2024 ROX Automation - Jev Kuznetsov
"""

import asyncio
import logging

import can

import rox_icu.can_protocol as canp
from rox_icu.can_utils import get_can_bus
from rox_icu.utils import run_main_async

NODE_ID = 0x01


log = logging.getLogger("master")


# Callback function to process received messages
def print_message(msg):
    log.info(f"callback: {msg}")


# Coroutine to send messages
async def send_messages(bus) -> None:
    io_state = 0

    try:
        for i in range(256):
            io_msg = canp.IOStateMessage(io_state)

            arb_id, data = canp.encode_message(io_msg, NODE_ID)

            bus_msg = can.Message(arbitration_id=arb_id, data=data)
            bus.send(bus_msg)

            # shift the io_state by 1 bit
            io_state = io_state << 1

            # reset the io_state to 1 if in overflow
            if io_state & 0xFF == 0:
                io_state = 1

            await asyncio.sleep(0.01)
        log.info("Finished sending messages")
    except asyncio.CancelledError:
        log.info("Send messages cancelled")
        raise


# Coroutine to receive messages
async def receive_messages(reader):
    try:
        while True:
            raw_msg = await reader.get_message()
            node_id = canp.get_node_id(raw_msg.arbitration_id)

            try:
                msg = canp.decode_message(raw_msg.arbitration_id, raw_msg.data)
            except KeyError:
                log.error(f"Unknown message: {raw_msg}")
                continue

            log.info(f"< node [{node_id}] message: {msg.__class__.__name__}")

    except asyncio.CancelledError:
        log.info("Receive messages cancelled")
        raise


async def main():
    # Start the sending and receiving coroutines
    bus = get_can_bus()

    reader = can.AsyncBufferedReader()
    notifier = can.Notifier(bus, [reader])

    try:
        # Create tasks for both coroutines
        send_task = asyncio.create_task(send_messages(bus))
        receive_task = asyncio.create_task(receive_messages(reader))

        # Wait for send_task to complete first
        await send_task

        # Once send_task is done, cancel the receive_task
        receive_task.cancel()
        try:
            await receive_task
        except asyncio.CancelledError:
            pass

        print("Done")

    except Exception as e:
        log.error(f"Exception: {e}")
        raise
    finally:
        log.info("Cleaning up")
        notifier.stop()  # Stop the notifier when done
        bus.shutdown()


run_main_async(main())



from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("rox_icu")
except PackageNotFoundError:  # pragma: no cover
    # Package is not installed, and therefore, version is unknown.
    __version__ = "0.0.0+unknown"



"""
Read inputs

Copyright (c) 2024 ROX Automation - Jev Kuznetsov
"""

import time
from icu_board import maxio, max1, max2, max_enable, D_PINS

maxio.DEBUG = False  # print debug info
max_enable.value = True  # enable in- and outputs

for pin in D_PINS:
    pin.switch_to_input()
    print(pin)


for drv in [max1, max2]:
    print("chip address:", drv.chip_address)

    drv.print_registers()

    # check pin configuration
    set_out = drv.read_register(maxio.REGISTERS.SET_OUT)
    print(f"SET_OUT: {set_out[1]:08b}")

# read inputs

while True:
    d_values = [int(pin.value) for pin in D_PINS]
    print(f"d values: {d_values}")
    time.sleep(0.5)



#!/usr/bin/env python3
"""
measure the time it takes to pack and parse a CAN message

reuslts (Feater M4):

---- before refactor to namedtuple ----
pack: 0.198 ms
parse: 0.174 ms

---- after refactor to namedtuple ----
nop: 0.005 ms
pack: 0.081 ms
parse: 0.097 ms
parse-direct: 0.079 ms


"""

import time
import gc
import struct
import can_protocol as protocol

opcode = protocol.get_opcode(protocol.HeartbeatMessage)
node_id = 10
iterations = 1000

byte_def = protocol.MESSAGES[opcode][1]

# --------nop---------
gc.collect()
t_start = time.monotonic_ns()
for _ in range(iterations):
    pass
t_elapsed = (time.monotonic_ns() - t_start) / 1e6
t_nop = t_elapsed / iterations
print(f"nop: {t_nop:.3f} ms")

# --------pack--------
gc.collect()
t_start = time.monotonic_ns()
for _ in range(iterations):
    msg = protocol.HeartbeatMessage(1, 2, 3, 4)
    data_bytes = struct.pack(byte_def, *msg)


t_end = time.monotonic_ns()
t_elapsed = (t_end - t_start) / 1e6
t_pack = t_elapsed / iterations
print(f"pack: {t_pack:.3f} ms")

# --------parse--------
gc.collect()
msg = protocol.HeartbeatMessage(1, 2, 3, 4)
msg_id = protocol.generate_message_id(node_id, opcode)

t_start = time.monotonic_ns()
for _ in range(iterations):
    msg2 = protocol.unpack(opcode, b"\x01\x02\x03\x04")

t_end = time.monotonic_ns()
t_elapsed = (t_end - t_start) / 1e6
t_parse = t_elapsed / iterations
print(f"parse: {t_parse:.3f} ms")


# --------parse direct  --------

gc.collect()

t_start = time.monotonic_ns()
for _ in range(iterations):
    msg2 = protocol.HeartbeatMessage(*struct.unpack(byte_def, b"\x01\x02\x03\x04"))

t_end = time.monotonic_ns()
t_elapsed = (t_end - t_start) / 1e6
t_parse = t_elapsed / iterations
print(f"parse-direct: {t_parse:.3f} ms")



#!/usr/bin/env python3
"""
ICU Mock CAN Interface

This mock simulates the ICU (Integrated Control Unit) CAN interface,
providing a virtual environment to test the Python driver for remote I/O
functionalities.


Copyright (c) 2024 ROX Automation - Jev Kuznetsov
"""

import asyncio
import logging

import can
from can.interfaces.socketcan import SocketcanBus
from can.interfaces.udp_multicast import UdpMulticastBus

import rox_icu.can_protocol as canp
from rox_icu.can_utils import get_can_bus
from rox_icu.utils import run_main

# Constants for CAN messages
NODE_ID = 0x01
DEVICE_TYPE = 201

# CAN logger to INFO level
logger_can = logging.getLogger("can")
if logger_can is not None:
    logger_can.setLevel(logging.INFO)


class ICUMock:
    """Class to mock the ICU CAN interface."""

    def __init__(
        self,
        node_id: int = NODE_ID,
        can_bus: SocketcanBus | UdpMulticastBus | None = None,
        simulate_inputs: bool = False,
    ):
        self._log = logging.getLogger(f"icu.mock.{node_id}")
        self.node_id = node_id

        self._io_state = 0  # Represents the state of all I/Os (8 bits for 8 I/Os)
        self.error_max1 = 0
        self.error_max2 = 0

        self._simulate_inputs = simulate_inputs

        self._bus = can_bus or get_can_bus()
        self._can_reader = can.AsyncBufferedReader()
        self._notifier = can.Notifier(self._bus, [self._can_reader])

    @property
    def io_state(self):
        """Get the IO state."""
        return self._io_state

    @io_state.setter
    def io_state(self, new_state: int) -> None:
        """Set the IO state."""
        self._log.info(f"Setting IO state: {new_state:02x}")
        self._io_state = new_state

        arb_id, data_bytes = canp.encode_message(
            canp.IOStateMessage(self._io_state), self.node_id
        )

        message = can.Message(
            arbitration_id=arb_id,
            data=data_bytes,
            is_extended_id=False,
        )
        self._bus.send(message)

    def get_global_error(self) -> tuple[int, int]:
        """Return the error status of max1 and max2."""
        return self.error_max1, self.error_max2

    async def message_handler(self) -> None:
        """Handle received CAN messages."""
        self._log.info("Starting message handler")

        while True:
            try:
                raw_msg = await self._can_reader.get_message()

                node_id = canp.get_node_id(raw_msg.arbitration_id)

                # Ignore messages that are not for this node ID
                if node_id != self.node_id:
                    continue

                self._log.debug(
                    f"Received message ID: {raw_msg.arbitration_id:x}, Data: {raw_msg.data.hex(' ')}"
                )

                msg = canp.decode_message(raw_msg.arbitration_id, raw_msg.data)

                if isinstance(msg, canp.IOSetMessage):
                    self._log.info(f"Received IOSetMessage: {msg.io_state:02x}")
                    self.io_state = msg.io_state

            except Exception as e:
                self._log.error(f"Error in message handler: {e}")

    async def heartbeat_loop(self, delay: float = 0.1) -> None:
        """Send heartbeat message."""
        self._log.info("Starting heartbeat loop")

        counter = 0

        while True:
            # Construct the heartbeat message
            heartbeat = canp.HeartbeatMessage(
                device_type=DEVICE_TYPE,
                error_max1=self.error_max1,
                error_max2=self.error_max2,
                io_state=self._io_state,
                errors=0,
                counter=counter,  # Increment counter for each loop
            )

            arb_id, data_bytes = canp.encode_message(heartbeat, self.node_id)

            message = can.Message(
                arbitration_id=arb_id,
                data=data_bytes,
                is_extended_id=False,
            )
            self._bus.send(message)
            counter += 1
            counter &= 0xFF  # Wrap around at 255
            await asyncio.sleep(delay)

    async def toggle_outputs(self):
        """Toggle the output pins."""
        if not self._simulate_inputs:
            return

        self._log.info("Starting output toggling loop")

        while True:

            # toggle bit 7
            self.io_state ^= 0x80

            await asyncio.sleep(0.5)

    async def main(self):
        """Main async loop for the ICU mock."""
        async with asyncio.TaskGroup() as tg:
            tg.create_task(self.heartbeat_loop())
            tg.create_task(self.message_handler())
            tg.create_task(self.toggle_outputs())

    def start(self):
        """Start the main loop."""
        asyncio.run(self.main())

    def __del__(self):
        """Destructor to clean up resources."""
        self._notifier.stop()
        self._bus.shutdown()


def main(node_id: int = NODE_ID):
    try:
        mock = ICUMock(node_id, simulate_inputs=True)
        mock.start()
    except KeyboardInterrupt:
        logging.info("KeyboardInterrupt - shutting down ICU mock")


if __name__ == "__main__":
    run_main(main)



"""

single loop test

Not the cleanest way to program, but most efficient.
Garbage collection however is dominant factor here, so async programming
should not be much slower than this.



"""

import time
import struct
import gc
import canio
from icu_board import can, maxio, max1, max2, max_enable
from micropython import const
import can_protocol as canp

NODE_ID = const(1)
MSG_ID = canp.generate_message_id(NODE_ID, canp.get_opcode(canp.HeartbeatMessage))

print(f"msg_id: {MSG_ID:x}")

# configuration
scope_pin = max2.d_pins[2]

# set scope pin to push-pull
max2.write_register(maxio.REGISTERS.CONFIG_DO, 0x30)

max_enable.value = True  # enable in- and outputs
counter = 0


d_pins = max1.d_pins + max2.d_pins

# set all D pins to input
for idx in range(4):
    max1.switch_to_input(idx)
    max2.switch_to_input(idx)

# set scope to output
max2.switch_to_output(2, False)

t_start = time.monotonic_ns()

t_loop_max = 0.0


hb_msg, hb_byte_def = canp.MESSAGES[0]

while True:
    # get state of all pins
    io_state = 0
    for bit, pin in enumerate(d_pins):
        io_state |= pin.value << bit

    # clear scope bit
    # io_state &= ~(1 << 6)

    # send heartbeat message
    msg = hb_msg(device_type=1, error_code=0, counter=counter & 0xFF, io_state=io_state)
    can_msg = canio.Message(id=MSG_ID, data=struct.pack(hb_byte_def, *msg))
    can.send(can_msg)

    # toggle scope pin
    scope_pin.value = not scope_pin.value

    # timing
    counter += 1

    t_now = time.monotonic_ns()
    t_loop = (t_now - t_start) / 1e6  # ms

    t_loop_max = max(t_loop_max, t_loop)

    if counter > 1000:
        print(f"loop time: {t_loop:.3f} ms, max: {t_loop_max:.3f} ms")
        t_loop_max = 0.0
        counter = 0

    t_start = t_now

    gc.collect()  # this takes quite some time



# pylint: disable=protected-access
import asyncio
import pytest

from rox_icu.core import Pin


def get_pin() -> Pin:
    return Pin(5)


def test_pin_initial_state():
    pin = get_pin()
    assert not pin.state
    assert pin.number == 5
    assert pin.high_event is not None
    assert pin.low_event is not None
    assert pin.change_event is not None


@pytest.mark.asyncio
async def test_pin_set_state():
    pin = get_pin()
    pin._update(True)
    assert not pin.is_input
    assert pin.state
    pin._update(False)
    assert not pin.state


@pytest.mark.asyncio
async def test_on_high_event():
    pin = get_pin()
    assert not pin.state

    # both events should be cleared
    assert not pin.low_event.is_set()
    assert not pin.high_event.is_set()

    # set low, no events should be triggered
    pin._update(False)
    assert not pin.low_event.is_set()
    assert not pin.high_event.is_set()

    # set high, on_high event should be triggered
    pin._update(True)
    assert not pin.low_event.is_set()
    assert pin.high_event.is_set()

    # check auto-clearing
    await pin.high_event.wait()
    assert not pin.high_event.is_set()  # AutoClearEvent should be cleared


@pytest.mark.asyncio
async def test_on_low_event():
    pin = get_pin()

    # both events should be cleared
    assert not pin.low_event.is_set()
    assert not pin.high_event.is_set()

    # set high, on_high should be triggered
    pin._update(True)
    assert not pin.low_event.is_set()
    assert pin.high_event.is_set()

    await pin.high_event.wait()
    assert not pin.high_event.is_set()

    # set low, on_low should be triggered
    pin._update(False)
    assert pin.low_event.is_set()
    assert not pin.high_event.is_set()

    # check auto-clearing
    await pin.low_event.wait()
    assert not pin.low_event.is_set()


@pytest.mark.asyncio
async def test_on_change() -> None:

    pin = get_pin()
    assert not pin.state

    async def toggle_state_loop(pin: Pin) -> None:
        """toggle state a couple of times"""

        for _ in range(10):
            pin._update(not pin.state)
            await asyncio.sleep(0.001)

    async def on_change_counter() -> int:
        """count on_change events with a timeout"""
        count = 0
        try:
            while True:
                await asyncio.wait_for(pin.change_event.wait(), timeout=0.1)
                count += 1
        except asyncio.TimeoutError:
            pass

        return count

    # Run both coroutines concurrently
    counter_task = asyncio.create_task(on_change_counter())
    toggle_task = asyncio.create_task(toggle_state_loop(pin))

    # Wait for toggle_task to complete first
    await toggle_task

    # Then wait for counter_task (it will finish when timeout occurs)
    count = await counter_task

    # Since we toggled 10 times, we should have 10 changes
    assert count == 10


# @pytest.mark.asyncio
# async def test_heartbeat() -> None:

#     node_id = 42
#     icu = ICU(node_id)

#     # no mock is running, so this should raise an error
#     with pytest.raises(HeartbeatError):
#         await icu.start()



import os
import time

from click import prompt
from invoke import task  # type: ignore

CWD = os.path.dirname(os.path.realpath(__file__))


def get_mount_point():
    mount_point = None
    for root, dirs, _ in os.walk("/media"):
        if "CIRCUITPY" in dirs:
            mount_point = os.path.join(root, "CIRCUITPY")
            print(f"CIRCUITPY mount point found: {mount_point}")
            return mount_point

    if mount_point is None:
        raise FileNotFoundError("CIRCUITPY mount point not found")


@task
def clean(ctx):
    """
    Remove all files and directories that are not under version control to ensure a pristine working environment.
    Use caution as this operation cannot be undone and might remove untracked files.

    """

    ctx.run("git clean -nfdx")

    if (
        prompt(
            "Are you sure you want to remove all untracked files? (y/n)", default="n"
        )
        == "y"
    ):
        ctx.run("git clean -fdx")


@task
def lint(ctx):
    """
    Perform static analysis on the source code to check for syntax errors and enforce style consistency.
    """
    # ctx.run("ruff check src tests")
    ctx.run("pylint -E src", warn=True, pty=True)
    ctx.run("pylint -E tests", warn=True, pty=True)
    ctx.run("mypy src", warn=True, pty=True)


@task
def test(ctx):
    """
    Run tests with coverage reporting to ensure code functionality and quality.
    """
    ctx.run("pytest --cov=src --cov-report term-missing tests", pty=True)


@task
def install(ctx):
    """install dependencies"""
    ctx.run("circup install -r cpy_requirements.txt")


@task
def sync(ctx):
    """sync code with the board"""
    mount_point = get_mount_point()

    # Directory where the lib files will be copied
    destination = os.path.join(mount_point, "lib")  # type: ignore

    # Ensure the destination directory exists
    os.makedirs(destination, exist_ok=True)

    # Copy files with rsync, following symlinks
    ctx.run(f"rsync -avL --exclude '_*' embedded_lib/ {destination}/")

    # Flush buffers
    ctx.run("sync")


@task(pre=[install, sync])
def init(ctx):
    """initialize the board, use blink as the main code"""
    mount_point = get_mount_point()

    # Copy the main code
    ctx.run(f"cp examples/101_hello_icu.py {mount_point}/code.py")

    # Flush


@task
def ci(ctx):
    """
    run ci locally in a fresh container

    """
    t_start = time.time()
    # get script directory

    try:
        ctx.run(f"docker run --rm -v {CWD}:/workspace roxauto/python-ci")
    finally:
        t_end = time.time()
        print(f"CI run took {t_end - t_start:.1f} seconds")


@task
def tox(ctx):
    """run tox tests in a fresh container"""
    ctx.run(
        f"docker run --rm -it -v {CWD}:/app divio/multi-python:latest tox", pty=True
    )


@task
def build_package(ctx):
    """
    Build package in docker container.
    """

    ctx.run("rm -rf dist")
    t_start = time.time()
    # get script directory
    script_dir = os.path.dirname(os.path.realpath(__file__))
    try:
        ctx.run(
            f"docker run --rm -v {script_dir}:/workspace roxauto/python-ci /scripts/build.sh"
        )
    finally:
        t_end = time.time()
        print(f"CI run took {t_end - t_start:.1f} seconds")


@task
def release(ctx):
    """publish package to pypi"""

    token = os.getenv("PYPI_TOKEN")
    if not token:
        raise ValueError("PYPI_TOKEN environment variable is not set")

    ctx.run(
        f"docker run --rm -e PYPI_TOKEN={token} -v {CWD}:/workspace roxauto/python-ci /scripts/publish.sh"
    )



#!/usr/bin/env python3
"""
 example of waiting for an edge on input and toggling an output with a delay

 Copyright (c) 2024 ROX Automation - Jev Kuznetsov
"""
import asyncio
import logging

from rox_icu.core import ICU
from rox_icu.utils import run_main_async

NODE_ID = 1

SENSOR_PIN = 7
RELAY_PIN = 0

log = logging.getLogger("main")


async def handle_input(icu: ICU) -> None:
    log.info("Starting input handler")

    sensor = icu.pins[SENSOR_PIN]
    relay = icu.pins[RELAY_PIN]
    relay.state = False

    while True:
        log.info("Waiting for high state")
        await sensor.high_event.wait()
        log.info(f"Sensor state: {sensor.state}")

        # toggle relay
        relay.state = True
        await asyncio.sleep(0.1)
        relay.state = False


async def main() -> None:

    icu = ICU(NODE_ID)
    await icu.start()

    try:
        await handle_input(icu)
    finally:
        await icu.stop()
        print("Done")


if __name__ == "__main__":
    run_main_async(main())



#!/usr/bin/env python3
"""
Read analog and display bar on the outputs.

Copyright (c) 2024 ROX Automation - Jev Kuznetsov
"""

import asyncio
import analogio
import math

from icu_board import rgb_led, led1, led2, D_PINS, max_enable, Pins

max_enable.value = True  # enable the MAX14906 chips


def pct_to_bar(pct: float) -> list:
    """convert percentage to a list of 8 elements for the bar display"""

    value = max(0, min(100, pct))

    # Calculate how many elements should be True
    true_count = round((value / 100) * 8)

    # Create the array
    return [1] * true_count + [0] * (8 - true_count)


async def flash_led(led, interval=0.5):
    """flash the LED at the given interval"""

    while True:
        led.value = not led.value
        await asyncio.sleep(interval)


async def glow_neopixel(period=5.0):
    """
    Make the Neopixel glow with a single color, varying brightness linearly

    :param period: Time in seconds for one complete brightness cycle (up and down)
    """
    print(f"Glowing the Neopixel with green color, {period}-second period")

    # Define the color (green)
    color = (0, 255, 0)  # Green

    # Set the pixel color
    rgb_led.fill(color)

    # Calculate the step time for smooth transition
    steps = 200  # Number of steps for a full cycle
    step_time = period / steps

    while True:
        for i in range(steps):
            # Use sine wave to smoothly transition between 0 and 1
            brightness = 0.5 * (1 + math.sin(2 * math.pi * i / steps - math.pi / 2))
            rgb_led.brightness = brightness
            await asyncio.sleep(step_time)


async def handle_analog() -> None:
    """read analog inputs and set nepixel colors"""
    print("Staring analog input handler")

    AIN1 = analogio.AnalogIn(Pins.A0)
    VREF = analogio.AnalogIn(Pins.VCC10V)

    c_offset = 500

    counter = 0

    while True:
        adc_val = AIN1.value
        adc_max_val = VREF.value
        pct = max(0, min(100, ((adc_val - c_offset) / (adc_max_val - c_offset)) * 100))

        bar_vals = pct_to_bar(pct)

        # set the D pins
        for idx, val in enumerate(bar_vals):
            D_PINS[idx].value = val

        counter += 1
        if counter % 100 == 0:
            led2.value = not led2.value
            print(f"{adc_val=} {pct=:.1f} {bar_vals=}")
        await asyncio.sleep(0)


async def main():
    """main coro"""
    await asyncio.gather(
        flash_led(led1),  # flash led1 with default frequency
        handle_analog(),  # read analog inputs and set neopixel colors
        glow_neopixel(),  # cycle through the RGB colors
    )


# ----------- run ---------------
asyncio.run(main())



# ECU board demo
import struct
import asyncio
import analogio
import canio
import gc

from icu_board import Pins, rgb_led, led1, led2, maxio, max1, max2, max_enable, can


maxio.DEBUG = False  # print debug info


button_pressed = asyncio.Event()

nr_button_presses = 0
nr_sensor_triggers = 0


def init_system() -> None:
    maxio.DEBUG = True

    rgb_led.brightness = 0.1
    rgb_led.fill((0, 0, 255))

    # set in- and outputs
    max1.switch_to_input(3)
    assert max1.read_register(maxio.REGISTERS.SET_OUT)[1] == 0x80

    max2.switch_to_input(0)
    assert max2.read_register(maxio.REGISTERS.SET_OUT)[1] == 0x10

    # maxio.write_register(maxio.REGISTERS.SET_OUT, 0x80, 0)
    # maxio.write_register(maxio.REGISTERS.SET_OUT, 0x10, 1)

    # read all max registers
    for drv in [max1, max2]:
        print("chip address:", drv.chip_address)
        drv.print_registers()

    print(48 * "-")

    # turn max on
    max_enable.value = True  # enable in- and outputs


async def flash_leds() -> None:
    while True:
        led1.value = not led1.value
        await asyncio.sleep(0.2)


async def read_button() -> None:
    prev_val = False

    global nr_button_presses
    button = max1.d_pins[3]

    while True:
        val = button.value
        if val != prev_val and val:
            nr_button_presses += 1
            print(f"Button pressed {nr_button_presses} times")
            button_pressed.set()
        prev_val = val
        await asyncio.sleep(0.01)


async def read_inductive_sensor() -> None:
    global nr_sensor_triggers

    prev_val = False

    output = max1.d_pins[0]

    while True:
        val = max2.d_pins[0].value
        output.value = val
        if val != prev_val and val:
            nr_sensor_triggers += 1
            print(f"Inductive sensor triggered {nr_sensor_triggers} times")
        prev_val = val
        await asyncio.sleep(0)


async def handle_analog() -> None:
    """read analog inputs and set nepixel colors"""

    AIN1 = analogio.AnalogIn(Pins.A0)
    AIN2 = analogio.AnalogIn(Pins.A1)

    c_offset = 700
    c_range = 47915 - c_offset

    rgb = [0, 0, 0]

    counter = 0

    while True:
        for idx, a in enumerate([AIN1, AIN2]):
            adc_val = a.value
            pct = max(0, min(100, ((adc_val - c_offset) / c_range) * 100))

            rgb[idx] = int(pct * 2.55)

        rgb_led.fill((rgb[0], 0, rgb[1]))
        counter += 1
        if counter % 10 == 0:
            led2.value = not led2.value
        await asyncio.sleep(0)


async def handle_outputs() -> None:
    """perform output actions"""

    output = max1.d_pins[1]  # attached led
    output.value = False

    relay = max1.d_pins[2]

    while True:
        await button_pressed.wait()
        print("Button pressed, toggling outputs")
        button_pressed.clear()

        # flash output
        for _ in range(10):
            output.value = not output.value
            await asyncio.sleep(0.1)

        output.value = False

        relay.value = True
        await asyncio.sleep(1.0)
        relay.value = False


async def send_can_messages() -> None:
    """send counter and number of button presses and sensor triggers over CAN"""
    counter = 0
    scope_pin = max2.d_pins[2]

    while True:
        message = canio.Message(
            id=0x01,
            data=struct.pack(
                "<BBB",
                counter & 0xFF,
                nr_button_presses & 0xFF,
                nr_sensor_triggers & 0xFF,
            ),
        )
        can.send(message)
        counter += 1
        scope_pin.value = not scope_pin.value
        gc.collect()
        await asyncio.sleep(0)


async def main() -> None:
    await asyncio.gather(
        flash_leds(),
        read_button(),
        handle_analog(),
        handle_outputs(),
        read_inductive_sensor(),
        send_can_messages(),
    )


# ---------main code----------


init_system()
asyncio.run(main())



#!/usr/bin/env python3
# Copyright (c) 2024 ROX Automation - Jev Kuznetsov
"""
ROX CAN Protocol Implementation

A lightweight CAN (Controller Area Network) protocol implementation for ROX devices,
compatible with both CPython and MicroPython for embedded systems use.

Protocol Structure
----------------
Message ID (11 bits):
    - Node ID: Upper 6 bits (0-63, node 0 reserved for broadcast)
    - Command ID: Lower 5 bits (0-31)

Message Types
------------
See the namedtuple definitions for message types.

Usage Example
-----------
>>> # Create and encode a message
>>> msg = HeartbeatMessage(device_type=1, error_max1=0, error_max2=0,
...                       io_state=1, device_state=0, counter=0)
>>> arb_id, data = encode_message(msg, node_id=1)
>>>
>>> # Decode received message
>>> decoded_msg = decode_message(arb_id, data)

Adding New Messages
-----------------
1. Define message structure using namedtuple
2. Add entry to _MSG_DEFS with (opcode, byte_definition)
3. Byte definition uses struct format chars (B=uint8, H=uint16, etc.)


"""
import struct
from collections import namedtuple

try:
    from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
    TYPE_CHECKING = False

if TYPE_CHECKING:
    from typing import Type, NamedTuple, Tuple  # pragma: no cover


VERSION = 10


class DeviceState:
    """Device state flags"""

    STOPPED = 0
    RUNNING = 1


# ----------------------------Message Definitions----------------------------
# opcode 0
HaltMessage = namedtuple("HaltMessage", "io_state")  # halt with desired io state.


# normally sent every 100ms
HeartbeatMessage = namedtuple(
    "HeartbeatMessage",
    ("device_type", "error_max1", "error_max2", "io_state", "errors", "counter"),
)

# sent on change
IOStateMessage = namedtuple("IOStateMessage", "io_state")

# requist to change io state
IOSetMessage = namedtuple("IOSetMessage", "io_state")


# ----------------------------Internal lookup tables---------------------------

# message: (opcode, byte_def)
_MSG_DEFS = {
    message_type: (opcode, byte_def)
    for opcode, (message_type, byte_def) in enumerate(
        [
            (HaltMessage, "<B"),
            (HeartbeatMessage, "<BBBBBB"),
            (IOStateMessage, "<B"),
            (IOSetMessage, "<B"),
        ]
    )
}
_OPCODE2MSG = {v[0]: k for k, v in _MSG_DEFS.items()}


# ----------------------------Utility Functions----------------------------
def generate_message_id(node_id: int, opcode: int) -> int:
    """Generates an 11-bit message ID from opcode and node ID."""
    return (node_id << 5) | opcode


def split_message_id(message_id: int) -> "tuple[int, int]":
    """Splits a 11-bit message ID into opcode and node ID."""
    opcode = message_id & 0x1F  # Extract lower 5 bits for cmd_id
    node_id = message_id >> 5  # Shift right by 5 bits to get node_id
    return node_id, opcode


def get_node_id(message_id: int) -> int:
    """Get the node ID from a message ID."""
    return message_id >> 5


def get_opcode_and_bytedef(message_class: "Type[NamedTuple]") -> "Tuple[int, str]":
    """Get the opcode for a message type."""
    return _MSG_DEFS[message_class]  # type: ignore


def encode_message(message: "NamedTuple", node_id: int) -> "Tuple[int, bytes]":
    """Pack a message into arbitration ID and data bytes.
    returns: (arbitration_id, data_bytes)"""
    opcode, byte_def = get_opcode_and_bytedef(type(message))
    arbitration_id = generate_message_id(node_id, opcode)
    return arbitration_id, struct.pack(byte_def, *message)


def decode_message(arb_id: int, data: "bytes | bytearray") -> "NamedTuple":
    """Parse a message from raw data."""
    opcode = arb_id & 0x1F
    message_class = _OPCODE2MSG[opcode]
    return message_class(*struct.unpack(_MSG_DEFS[message_class][1], data))



"""
Toggle outputs

Copyright (c) 2024 ROX Automation - Jev Kuznetsov
"""

import time
from icu_board import maxio, max1, max2, max_enable, D_PINS


maxio.DEBUG = True  # print debug info
max_enable.value = True  # enable in- and outputs

for drv in [max1, max2]:
    print("chip address:", drv.chip_address)
    error = drv.get_global_error()
    print("global error:", error)

    drv.print_registers()

    # toggle outputs with D pins

try:
    while True:
        for pin in D_PINS:
            print("toggling", pin)
            pin.value = True
            time.sleep(0.5)
            pin.value = False
finally:
    # disable all outputs
    max_enable.value = False



"""
Running ICU as a remote IO device.



"""

import asyncio
import struct
import canio
from icu_board import led1, can, maxio, max1, max2, max_enable, D_PINS
from micropython import const
import can_protocol as canp

NODE_ID = const(0x01)

BUTTON = const(7)
SENSOR = const(0)

# -------------- initialisation ----------------
print(f"Can protocol version: {canp.VERSION}")

# intialize system
max_enable.value = True  # enable in- and outputs

# configure inputs
for pin_nr in [BUTTON, SENSOR]:
    D_PINS[pin_nr].switch_to_input()

# show state of all d pins
for pin in D_PINS:
    print(pin)


def get_io_state() -> int:
    io_state = 0
    for bit, pin in enumerate(D_PINS):
        io_state |= pin.value << bit
    return io_state


async def read_inputs() -> None:
    """read inputs and send can message on change"""

    # configuration

    # set scope pin to push-pull
    max2.write_register(maxio.REGISTERS.CONFIG_DO, 0x30)

    io_state = get_io_state()
    prev_io_state = io_state

    opcode, byte_def = canp.get_opcode_and_bytedef(canp.IOStateMessage)
    msg_id = canp.generate_message_id(NODE_ID, opcode)
    print(f"IOStateMessage ID: {msg_id:x}")

    while True:
        # get state of all pins
        io_state = get_io_state()

        if io_state != prev_io_state:
            msg = canio.Message(id=msg_id, data=struct.pack(byte_def, io_state))
            can.send(msg)

        prev_io_state = io_state

        await asyncio.sleep(0)


async def heartbeat_loop() -> None:
    counter = 0

    opcode, byte_def = canp.get_opcode_and_bytedef(canp.HeartbeatMessage)

    msg_id = canp.generate_message_id(NODE_ID, opcode)
    print(f"HeartbeatMessage ID: {msg_id:x}")

    while True:
        msg = canp.HeartbeatMessage(
            device_type=1,
            error_max1=max1.get_global_error(),
            error_max2=max2.get_global_error(),
            io_state=get_io_state(),
            device_state=0,
            counter=counter & 0xFF,
        )
        can_msg = canio.Message(id=msg_id, data=struct.pack(byte_def, *msg))
        can.send(can_msg)

        led1.value = not led1.value

        counter += 1

        await asyncio.sleep(0.1)


async def main() -> None:
    await asyncio.gather(read_inputs(), heartbeat_loop())


asyncio.run(main())



import struct

import pytest

from rox_icu import can_protocol as canp

EXPECTED_PROTOCOL_VERSION = 10


def test_version() -> None:
    assert canp.VERSION >= EXPECTED_PROTOCOL_VERSION


def test_split() -> None:
    message_id = 0x2C

    assert canp.split_message_id(message_id) == (0x01, 0x0C)


def test_node_id() -> None:
    for node_id in range(64):
        assert canp.get_node_id(canp.generate_message_id(node_id, 0)) == node_id


def test_roundtip() -> None:
    for endpoint in range(32):
        for node_id in range(64):
            message_id = canp.generate_message_id(node_id, endpoint)
            assert canp.split_message_id(message_id) == (node_id, endpoint)


def test_invalid_message() -> None:
    class InvalidMessage:
        pass

    with pytest.raises(KeyError):
        canp.get_opcode_and_bytedef(InvalidMessage)  # type: ignore


def test_halt() -> None:
    test_bytes = b"\x01"

    msg = canp.HaltMessage(1)
    assert msg.io_state == 1

    opcode, byte_def = canp.get_opcode_and_bytedef(canp.HaltMessage)
    assert opcode == 0
    assert byte_def == "<B"

    data_bytes = struct.pack(byte_def, *msg)

    assert data_bytes == test_bytes

    # use function
    arb_id, data_bytes = canp.encode_message(msg, 1)
    assert data_bytes == test_bytes

    # convert back
    msg2 = canp.HaltMessage(*struct.unpack(byte_def, test_bytes))

    assert msg == msg2


def test_heartbeat() -> None:
    test_bytes = b"\x01\x02\x03\xbe\xef\xff"

    msg = canp.HeartbeatMessage(1, 2, 3, 0xBE, 0xEF, 0xFF)
    assert msg.device_type == 1
    assert msg.error_max1 == 2
    assert msg.error_max2 == 3
    assert msg.io_state == 0xBE
    assert msg.errors == 0xEF
    assert msg.counter == 0xFF

    opcode, byte_def = canp.get_opcode_and_bytedef(canp.HeartbeatMessage)
    assert opcode == 1
    assert byte_def == "<BBBBBB"

    data_bytes = struct.pack(byte_def, *msg)
    assert data_bytes == test_bytes

    # # convert back
    msg2 = canp.HeartbeatMessage(*struct.unpack(byte_def, test_bytes))

    assert msg == msg2

    # use function
    _, data_bytes = canp.encode_message(msg, 1)
    assert data_bytes == test_bytes


def test_pack_unpack() -> None:
    test_bytes = b"\x01\x02\x03\xbe\xef\xff"

    msg = canp.HeartbeatMessage(1, 2, 3, 0xBE, 0xEF, 0xFF)

    msg_id, data_bytes = canp.encode_message(msg, 1)

    assert data_bytes == test_bytes

    # convert back
    msg2 = canp.decode_message(msg_id, data_bytes)
    assert msg == msg2
